# Xtractor.py
# FAB LIBRARY ROUTINE
# Fetch a URL and output the n most representative stemmed keywords

# This version 5.1
# Last update 14 Sept 1995

# Copyright 1995 The Board of Trustees of The Leland Stanford
# Junior University. All Rights Reserved

# Version History
# v5.1 created 14 Sept 1995
#      Added support for redirects
#      (changes output of main function)
# v5.0 created 7 July 1995

# -----------------------------------------------------------------
# Imports

import sys, string, regex, regsub, marshal, os
import proxy_urllib
from math import log, sqrt

# HACK
try:
	from stem import getstem
except:
	def getstem(x): return x

from HttpUtils import *
from general import *
from FabGlobals import *
import RobotsLib

# -----------------------------------------------------------------
# Weed the crap out of the last added link
# (assumes the URL is the first thing in double quotes
#  after the href)

def WeedLinks(Links,LastIndex,ContextUrlName):

	Link = Links[LastIndex]
	try:
		GlobalisedUrl = WeedLink(Link,ContextUrlName)

		EscapedUrl = EscapeQuotes(GlobalisedUrl)
		
		# Remove a #index if there is one
		StrippedUrl = proxy_urllib.splittag(EscapedUrl)[0]
		if StrippedUrl == None:
			StrippedUrl = ''

		if StrippedUrl == '':

			# Empty link!
			# error('Empty link found while parsing '+ContextUrlName)
			Links = Links[:-1]
			LastIndex = LastIndex - 1

		elif proxy_urllib.splittype(StrippedUrl)[0] == 'mailto':

			# Ignore mailto links
			Links = Links[:-1]
			LastIndex = LastIndex - 1

		elif StrippedUrl in Links[:LastIndex]:
			
			# Got this link already, ignore it
			Links = Links[:-1]
			LastIndex = LastIndex - 1

		else:

			Links[LastIndex] = StrippedUrl

	except ValueError,IndexError:
		# If a URL did not parse, remove it from the end of the list
		Links = Links[:-1]
		LastIndex = LastIndex - 1

	return (Links,LastIndex)

# -----------------------------------------------------------------
# Given the text of an HTML document with URL ContextUrlName,
# return a tuple (Words,Links) where Words is a dictionary of all
# the non-HTML words with more than 2 characters, with a count
# for each one,
# and Links is a list of the globalised links (not including inline
# images)

def ParseHTML(Text,ContextUrlName,Stops):

	# Split into words separated by white space
	Words = string.split(Text)

	NewWords = {}
	Links = []
	InLink = FALSE
	LinkNum = -1
	WordNum = -1

	crap = regex.compile('.*<.*>.*\|<.*\|.*>\|.*@.*\|.*:/.*\|.*=.*\|.*href.*')
	
	for Word in Words:
		
		WordNum = WordNum + 1

		# Space out HTML tags so they will be separate words
		fend = len(Word)-1
		f1 = string.find(Word,'<',1)
		if (0 < f1 < fend):
			Words.insert(WordNum+1,Word[f1:])
			Word = Word[:f1]
		f2 = string.find(Word,'>',1)
		if (0 < f2 < fend):
			Words.insert(WordNum+1,Word[(f2+1):])
			Word = Word[:(f2+1)]
		
		if Word == '':
			# Ignore empty words
			continue
		else:
			LoweredWord = string.lower(Word)

		# WORD EXTRACTING
		# ---------------
		# Remove HTML tag words or URL's or email addresses

		try:
			CrapMatch = crap.match(LoweredWord)
		except:
			# This will fail on a stack overflow
			# or something like that

			WeirdError('regex match failure for '+LoweredWord)

			# In that case ignore this word
			CrapMatch = 1

		if CrapMatch == -1:
			# Make word letters only
			W1 = ToAlpha(LoweredWord)			

			# Ignore words less than 3 chars long and
			# stop words
			if W1 not in Stops:

					# Stemming
					W2 = getstem(W1)

					if len(W2) > 2:

						# Update dictionary of words found
						try:
							NewWords[W2] = NewWords[W2] + 1
						except KeyError:
							# Not already in dictionary
							NewWords[W2] = 1
		
		# LINK EXTRACTING
		# ---------------
		# If the word is the beginning of a link, start extracting
		if LoweredWord == '<a':
			InLink = TRUE
			Links.append('')
			LinkNum = LinkNum + 1
		elif LoweredWord[-1] == '>':
			# Stop extracting when you see a >
			if InLink:
				Links[LinkNum] = Links[LinkNum] + Word + ' '
				(Links,LinkNum) = WeedLinks(Links,LinkNum,ContextUrlName)
			InLink = FALSE

		if InLink:
			Links[LinkNum] = Links[LinkNum] + Word + ' '

	return (NewWords,Links)

# -----------------------------------------------------------------	

def CalculateWeights(Words,D,Dlen,DocNum):
	
	if Words == {}:
		return {}
	Weights = {}
	WordsKeys = Words.keys()
	WordsValues = Words.values()
	MaxFreq = float(max(WordsValues))  
	#DocNum = float(max(D.values()))    # KLUDGE!
	DocNum = float(DocNum)

	# Work out TFIDF for each term
	for i in range(0,len(Words)-1):
		Word = WordsKeys[i]
		Freq = WordsValues[i]
		try:
			DocFreq = D[Word]
		except KeyError:
			# That word isn't in the dictionary
			DocFreq = 1

		# Term Freq normalised to lie between 0.5 and 1
		TF = 0.5 + (0.5 * (Freq / MaxFreq))

		# NB using base 10 log for Inverse Doc Frequency
		IDF = log(DocNum / DocFreq) 

		Weights[Word] = TF * IDF

	return Weights

# -----------------------------------------------------------------
def NormaliseWeights(Weights):

	# Work out normalisation factor NORM,
	# just the euclidean length of the vector
	W1 = map( lambda x: x*x, Weights)
	Sum = 0
	for i in W1:
		Sum = Sum + i
	NORM = sqrt(Sum)

	if NORM == 0:
		# Eeek! A zero length vector?
		# This is weird
		# Kludge this for now
		# (All weights will be 0 anyway so
		#  what we set NORM to doesnt matter)
		NORM = 1

	# Now replace TFIDF term weights by TFIDF/NORM
	# (cant use NORM inside lambda so need to pass
	#  it in as a fake parameter)
	return map( lambda x,n=NORM: x/n, Weights)


# -----------------------------------------------------------------
# Find the top N words in the URL UrlName
# Given D as a dictionary containing a count of all words seen so far
# and DocNum = the number of documents seen so far

# (If we already have the text can pass it in)

def URL2words(D,Dlen,DocNum,UrlName,N,Stops,h,Text=None):

	if not Text:
		NewUrl,UrlContentType,Text = h.FetchUrl2(UrlName)
		if NewUrl: UrlName = NewUrl

	(Words,Links) = ParseHTML(Text,UrlName,Stops)

	Title = FindTitle(Text)
	Site = GetSite(UrlName)
	
	# Calculate Weights
	Weights = CalculateWeights(Words,D,Dlen,DocNum)

	# Need to reverse order of items in dictionary so that
	# weight is first item in list so that we can sort on it
	WeightsSorted = map(lambda (x,y): (y,x),Weights.items())

	WeightsSorted.sort()
	# Sorted so that highest-weight words are at the end

	FinalWords = []
	FinalWeights = []

	# If there are less than N words, just output the ones we have
	for i in range(len(Weights)-1,max(len(Weights)-N-1,-1),-1):
		FinalWords.append(WeightsSorted[i][1])
		FinalWeights.append(WeightsSorted[i][0])

	# Finally normalise weights so length of vector = 1
	NormWeights = NormaliseWeights(FinalWeights)
		
	return (FinalWords,NormWeights,Links,Site,Title,UrlName)


# -----------------------------------------------------------------
# Import a dictionary and DocNum from disk

def LoadDict(FileName):
	try:
		f = open(FileName,'r')
		DD = marshal.load(f)
		f.close()
	except IOError:
		# error('Could not read dictionary file '+FileName)
		return ({},1)   # 0 causes lg to barf later on
	return DD		

# -----------------------------------------------------------------
# Import Stop Words List

def LoadStops(FileName):

	try:
		f = open(FileName,'r')
		Words = marshal.load(f)
		f.close()
	except IOError:
		# error('Could not read stop words file '+FileName)
		return([])
	return Words


# -----------------------------------------------------------------
# Import Bad Sites Table

# def LoadBadSites(FileName):

# 	try:
# 		f = open(FileName,'r')
# 		BadSites = marshal.load(f)
# 		f.close()
# 	except IOError:
# 		# error('Could not read bad sites file '+FileName)
# 		return({})
# 	return BadSites


# -----------------------------------------------------------------

def FilterByDomain(Links,AcceptableDomains,BadSites={}):

	GoodLinks = []
	for l in Links:

		try:
			type,rest = proxy_urllib.splittype(l)
			hostport,rest = proxy_urllib.splithost(rest)
			host,port = proxy_urllib.splitport(hostport)
			domain = os.path.splitext(host)[1][1:]
		except:
			# Could not get hold of domain at all!
			# Ditch this link
			continue

		if (domain in AcceptableDomains) or (AcceptableDomains == []):
			if not BadSites.has_key(host):
				GoodLinks.append(l)

	return GoodLinks


# -----------------------------------------------------------------
# EXTERNAL FUNCTIONS

class Xtractor:

	# Intialisation
	def __init__(self,L=NullLogger()):
	
		self.L = L

		# Initialise Http Utilities
		self.h = HttpUtils()
		self.h.SetAcceptableContentTypes(ACCEPTABLE_CONTENT_TYPES)
		self.h.SetProxy(PROXY)
		self.h.SetTimeout(TIMEOUT)
		self.h.SetMaxbytes(MAXBYTES)

		# New exception
		self.ContentTypeWrong = self.h.ContentTypeWrong

		# Import dictionary
		(self.D,self.DocNum) = LoadDict(DICT_FILENAME)
		self.Dlen = len(self.D)

		# Import stop words
		self.Stops = LoadStops(STOPS_FILENAME)


		# Import bad sites list
		#self.BadSites = LoadBadSites(BAD_SITES_FILENAME)
		# New version - get it from a global variable
		self.BadSites = EXTRA_BAD_SITES

		# Initialise the robots.txt checking
		self.R = RobotsLib.RobotsCheck(self.L)

		# Only return child links if they have one of
		# these domains (None means accept all domains)
		self.AcceptableDomains = ACCEPTABLE_DOMAINS


	# Call for one URL
	def XtractUrl(self,UrlName,N=NUM_WORDS):
		# Find keywords and links
		# This might raise IOError or self.ContentTypeWrong

		if self.R.CheckSite(UrlName):
			(Words,Weights,Links,Site,Title,Url) = URL2words(self.D,self.Dlen,self.DocNum,UrlName,N,self.Stops,self.h)
		else:
			raise IOError,'Disallowed by robots.txt'

		if self.AcceptableDomains or self.BadSites:
			Links = FilterByDomain(Links,self.AcceptableDomains,self.BadSites)

		return (Words,Weights,Links,Site,Title,Url)


	# Return all the stemmed words and their frequencies
	# (as a dictionary)
	# for use in compiling IDF dictionaries
	def GetWords(self,UrlName):
		
		# Fetch the URL
		NewUrl,UrlContentType,Text = self.h.FetchUrl2(UrlName)
		if NewUrl: UrlName = NewUrl

		# Extract the words
		(Words,Links) = ParseHTML(Text,UrlName,self.Stops)
		
		return Words


	# Tidy Up
	def __del__(self):
		self.close()
	def close(self):
		# Make sure this is only run once
		if hasattr(self,'Stops'):
			del self.Stops
			del self.D
			del self.DocNum
			del self.h
			self.R.close()












